#ifndef lint
static char SCCSidI[] = "@(#) ./comm/initeui.c 12/01/93";
#endif

/*
    IBM-EUI specific routines.

    An assumption here is that the C-interface will be natural AND
    compatible with programs that also use Fortran.  If this is not
    true, a lot of routines will require parameters to be passed
    as addresses....
    
 */
int  __EUIFROM, __EUILEN, __EUITYPE;
int  __NUMNODES, __MYPROCID;
int  __ALLGRP;
int  __EUIBSendLen;

void PIEUI_BSendInit();

/*
    euiexitall - terminate all processes in an execution (parallel exit)
    See init.c for a complete description of the interface
 */
static void euiexitall(msg,rc)
char *msg;
int  rc;
{
if (msg) fprintf( stderr, msg );
#if defined(USE_EUI_FORTRAN_INTERFACE)
mp_stopall( &rc );
#else
mpc_stopall( rc );
#endif
}

#define PI_HAS_PIINIT
void PIinit()
{
int min, max;
int rc, nbuf[4];
#if defined(USE_EUI_FORTRAN_INTERFACE)
int three = 3, four = 4;
mp_environ(  &__NUMNODES, &__MYPROCID );
rc = mp_task_query( nbuf, &four, &three );
#else
mpc_environ( &__NUMNODES, &__MYPROCID );
rc = mpc_task_query( nbuf, 4, 3 );
#endif
__ALLGRP = nbuf[3];

/* Close to maxint; this is the threshold where we switch to
   nonblocking messages */
__EUIBSendLen = 0x7ffffff;
PIEUI_BSendInit();

PISetupCollectiveTree( );
/* Default mesh is 1-d in x */
PISetMeshSize( __NUMNODES, 1 );
PISetNbrRoutine( PIGetNbrsMSH );

/* Make sure that the error handling routines can detect out-of-range tags */
PITagRange( &min, &max );
PIiSetTagLimits( min, max );
}

#include <signal.h>

/*
   Horribly enough, this program needs to do a exec with the value of 
   MP_PROCS set to the value of -np.  Further, there is NO way to 
   select the nodes to use.

   Rumor has it that these restrictions will change.  We'll see...
 */
int PIcall( np, procgroup, pfname, routine, argc, argv )
int  np, argc;
char **procgroup, *pfname;
char **argv;
int  (*routine)();
{
int rc, fsec;

/* The first thing that we do for EUI-H programs is to re-exec the appropriate
   controller... */
#define PI_NICE_STARTUP
#if defined(PI_NICE_STARTUP) && defined(euih)
PIiEUIHRun( &argc, argv );
#endif


/* EUI seems to catch SIGINT, we catch rest here */
/* Reason: it seems SIGSEGV don't get passed back to host */
/*         so job may hang. We want parallel job to end by default */
/*         user may change in worker() routine */
SYSetDefaultSignals( (void (*)())0 );
/* EUI seems to send SIGTSTP, SIGCONT and SIGTTIN to nodes we want them ignored */
/* because when they receive SIGCONT or SIGTTIN they crash */
signal(SIGTSTP,SIG_IGN);
signal(SIGCONT,SIG_IGN);
/* This DOES NOT WORK; if host job is put in background and then keyboard used */
/* one gets process_io: stdin error: Interrupted system call */
/* they claim this was fixed in next release */
signal(SIGTTIN,SIG_IGN);

SYSetExitAll(euiexitall);

/* 
   IBM claims that the DONTCARE value is a variable.  But they always
   have been using -1.  We use -1 for simplicity and efficiency, and
   test that assumption here
 */
   { int one = 1, three = 3, nbuf[1];
   mp_task_query( nbuf, &one, &three );
   if (nbuf[0] != -1) {
	fprintf( stderr, 
	  "Chameleon EUI implementation found different DONTCARE (=%d)\n", 
		nbuf[0] );
	SYexitall((char *)0,1);
        }
   }

/* Set resource limits */
SYGetResourceDefaults( &cpu, &mem, &pf, &niceval, &fsec );
/* Defaults are too small for SP so we do temporary ugly fix */
if (cpu == 120 && mem == 8 && fsec == 300) { 
    cpu = 600; mem = 64;  fsec = 1200;}
SYGetDayTime( &maxtime ); maxtime.tv_sec += fsec;
SYGetResourceLimits( &argc, argv, &cpu, &mem, &pf, &maxtime, 1 );
SYSetResourceLimits( mem, cpu, pf, &maxtime );
/* Turn off the resource clock for EUI/EUI-H */
SYSetResourceClockOff();
if (niceval > 0) SYSetNice( niceval );

PIinit();
#if !defined(LOGCOMMDISABLE)
PIGetDebugArgs( &argc, argv, __MYPROCID == 0 );
#endif

if (SYArgHasName( &argc, argv, 1, "-listnodes" )) {
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif
    char name[MAXHOSTNAMELEN];
    SYGetHostName( name, MAXHOSTNAMELEN );
    printf( "%s\n", name );
    }

/* Check for p4-specific options */
{
char cbuf[20];
while (SYArgGetString( &argc, argv, 1, "-eui", cbuf, 20 ))
    PISetOption( "eui", cbuf, (void *)0 );
}

PIinitlog();
PISummaryClear();
rc = (*routine)( argc, argv );
PIendlog();
{
int cnt;
if (cnt = PIEUI_BSendFlush( )) {
    fprintf( stderr, "[%d] found %d uncompleted blocking sends\n", 
	     PImytid, cnt );
    }
}

return rc;
}

/* Routine that can not easily be implemented as a macro */
int mppi_asyncprobe( type )
int type;
{
__EUITYPE = type;
__EUIFROM = DONTCARE;

/* Note probe supported by EUI-H.  Using this means that the library is
   incompatible with EUI without also providing a mp_probe stub that 
   calls euiexitall.

   Note: mp_probe now available with EUI.  It appears to work.
*/
/* #if defined(euih) */
#if defined(USE_EUI_FORTRAN_INTERFACE)
mp_probe( &__EUIFROM, &__EUITYPE, &__EUILEN );
#else
mpc_probe( &__EUIFROM, &__EUITYPE, &__EUILEN );
#endif
return __EUILEN >= 0;

/*
 #else
 euiexitall( "probe is not supported by EUI-1", 1);
 #endif
 */
}

/*
  There are times when we want to not worry about having blocking sends
  block because of resource exhaustion on the receiver (for example, 
  no matching recieve posted before the receiver did its own send.

  This (optional) code works by changing all blocking sends into non-blocking
  sends
 */

typedef struct _Pending {
    char *buf;
    int  sid;
    struct _Pending *next;
    } Pending;

/* We preallocate some send blocks; for now, if this is exceeded, the code
   will fall back on blocking sends */

#define PINSENDSAVAIL 128
static Pending pnd[PINSENDSAVAIL];
static Pending *head = 0, *avail;

/* Number of bytes for where we switch to the non-blocking sends */
#define PIBSENDTHRESHOLD __EUIBSendLen

/* Initialize the initial structures */
void PIEUI_BSendInit()
{
int     i;

for (i=0; i<PINSENDSAVAIL-1; i++) 
    pnd[i].next = pnd + i + 1;
pnd[PINSENDSAVAIL-1].next = 0;
avail = pnd;
head  = 0;
}

void PIEUI_BSend( tag, buffer, len, to, datatype )
int  tag, len, to, datatype;
char *buffer;
{
Pending *new, *prev, *next;

if (head) {
    /* Check for completed sends */
    new  = head;
    prev = 0;
    while (new) {
	next = new->next;
	if (mp_status( &new->sid ) >= 0) {
	    if (prev) {
		prev->next = next;
		}
	    else {
		head       = next;
		}
	    new->next = avail;
	    avail     = new;
	    FREE( new->buf );
	    new = next;
	    }
	else {
	    prev = new;
	    new  = next;
	    }
	}
    }
/* fprintf( stderr, "threshold is %d\n", __EUIBSendLen ); fflush( stderr ); */
if (len >= PIBSENDTHRESHOLD && avail) {
    /* fprintf( stderr, "Sending message using nonblocking send...\n" ); */
    /* fflush( stderr ); */
    new      = avail;
    avail    = avail->next;
    new->buf = (char *)MALLOC( len );   CHKPTR(new->buf);
    memcpy(new->buf, buffer, len );
    mp_send( new->buf, &len, &to, &tag, &new->sid );
    new->next = head;
    head      = new;
    return;
    }

mp_bsend( buffer, &len, &to, &tag );
}

/* Checks for unsatisfies non-blocking receives and returns the number of
   such receives */
int PIEUI_BSendFlush( )
{
Pending *new, *prev, *next;
int     cnt;

cnt = 0;
if (head) {
    /* Check for completed sends */
    new  = head;
    prev = 0;
    while (new) {
	next = new->next;
	if (mp_status( &new->sid )) {
	    if (prev) {
		prev->next = next;
		}
	    else {
		head       = next;
		}
	    new->next = avail;
	    avail     = new->next;
	    FREE( new->buf );
	    new = next;
	    }
	else {
	    cnt++;
	    prev = new;
	    new  = next;
	    }
	}
    }
return cnt;
}

#ifndef __MSDOS__

#include <fcntl.h>

#endif

#define PI_HAS_SET_OPTION
void PISetOption( version, name, val )
char *version, *name;
void *val;
{
if (strcmp( version, "eui" ) != 0) return;

if (strcmp( name, "list" ) == 0) {
    if (!val) val = (void *)stdout;
    fprintf( (FILE*)val, "list  - generate this list\n" );
#if defined(euih)
    fprintf( (FILE*)val, 
"nsend=nn - replace blocking with non-blocking sends for messages\n\
longer than nn bytes\n" );
    fprintf( (FILE*)val, 
"stdin=filename - use filename as standard input\n" );
#endif
    return;
    }

if (strncmp( name, "nsend", 5 ) == 0) {
    __EUIBSendLen = atoi( name + 6 );
    }
#if defined(euih)
if (strncmp( name, "stdin", 5 ) == 0) {
    /* EUI-H has no way to access stdin, so we handle that here. */
    int fd1, fd2, fd3;
    fd1 = fileno( stdin );
    fd2 = open( name+6, O_RDONLY );
    fd3 = dup2( fd2, fd1 );
    /* fd3 and fd1 should no be the same */
    close( fd2 );
    }
#endif
}

#define PI_HAS_EXIT
#define PI_HAS_INIT

