/*
   This is a test program for the fd write (and later read) routines

   This will eventually be expanded to automatically experiment with
   buffer sizes 
 */

#include "tools.h"
#include "system/system.h"
#include "comm/comm.h"
#include "comm/io/pio.h"
#include <math.h>

/* #define DBG(str) {fprintf( stdout, "[%d]%s\n", PImytid, str);fflush(stdout);} */
#define DBG(str)
int main( argc, argv )
int  argc;
char **argv;
{
int worker();
PICall( worker, argc, argv );
}

int worker( argc, argv )
int  argc;
char **argv;
{
int         nx, ny, nr, nc, myid, ri, rj, i, j, g;
int         nxin, nyin, gin;
int         px, py;    /* Size of partition in each dimension */
PIFILE      *fp;
PIArrayPart sz[2];
int         nd;
double      *v, *v2;
double      t1, t2;
int         bytes;
int         rsize = 0, wsize = 0;
char        fname[1000];
char        *fmat, *frmat;
int         flen;
int         err;
int         pmode;

strcpy( fname, "foo" );
/* get the global size */
if (SYArgHasName( &argc, argv, 1, "-help" )) {
    fprintf( stderr, 
	    "%s -n nx -m ny -g n -rsize n -wsize n -f name -noformat\n\
                -block -pio\n", 
	    argv[0] );
    return 0;
    }
nx = 16;
ny = 16;
g  = 0;
SYArgGetString( &argc, argv, 1, "-f", fname, 1000 );
SYArgGetInt( &argc, argv, 1, "-n", &nx );
SYArgGetInt( &argc, argv, 1, "-m", &ny );
SYArgGetInt( &argc, argv, 1, "-g", &g );
SYArgGetInt( &argc, argv, 1, "-rsize", &rsize );
SYArgGetInt( &argc, argv, 1, "-wsize", &wsize );
/* Block switches to the "buffered version"; don't use for AS_PARALLEL */
if (SYArgHasName( &argc, argv, 1, "-block" ))
     PIFSetWritersBlock();
fmat  = "%12.4e\n";
frmat = "%12le\n";
flen  = 13;
if (SYArgHasName( &argc, argv, 1, "-noformat" )) {
    fmat  = 0;
    frmat = 0;
    flen  = sizeof(double);
    }
pmode = PIO_AS_SEQUENTIAL;
if (SYArgHasName( &argc, argv, 1, "-pio" )) pmode = PIO_AS_PARALLEL;
fp = PIFopen( fname, PIAllProcs, O_WRONLY | O_CREAT, pmode, -1 );
/* Switch to cached */
if (rsize > 0 || wsize > 0) 
    PIFCacheEnable( fp, rsize, wsize );


PIWriteCommon( fp, "%4d",   4, &nx, 1, MSG_INT );
PIWriteCommon( fp, "%4d",   4, &ny, 1, MSG_INT );
PIWriteCommon( fp, "%4d\n", 5, &g,  1, MSG_INT );
/* Double align so that viewing the unformatted part of the file is easier */
if (!fmat)
    PIWriteCommon( fp, "  \n", 3, (void *)0, 1, MSG_INT );
nd   = 2;
nr   = PInumtids;
myid = PImytid;
if (nr > 1) 
    nr = sqrt( (double)nr + 0.1 );
nc = PInumtids / nr;
if (nr * nc != PInumtids) {
    fprintf( stderr, "Test only works for square arrays of processors\n" );
    return;
    }
ri = myid % nc;
rj = myid / nc;
    
px = nx / nc;
py = ny / nr;
/* printf( "[%d](%d,%d)\n", PImytid, ri, rj ); */
/* 
   The GLOBAL array is A(0:nx-1,0:ny-1).  The local array is
   declared as B(0:,0:), with B(g:,g:) matching A(ri*(nx/nc):,rj*(ny/nr):), 
   where the position of the processor in the mesh is given by (ri,rj)
 */
sz[0].mdim   = nx;
sz[0].ndim   = px + 2 * g;
sz[0].start  = g;
sz[0].end    = px - 1 + g;
sz[0].gstart = ri * px;
sz[0].gend   = sz[0].gstart + sz[0].ndim - 1;

sz[1].mdim   = ny;
sz[1].ndim   = py + 2 * g;
sz[1].start  = g;
sz[1].end    = py - 1 + g;
sz[1].gstart = rj * py;
sz[1].gend   = sz[1].gstart + sz[1].ndim - 1;
/* printf( "[%d]<%d,%d>\n", PImytid, sz[0].gstart, sz[1].gstart );  */
fflush(stdout);
v = (double *)MALLOC( sz[0].ndim * sz[1].ndim * sizeof(double) );  CHKPTR(v);
for (j=sz[1].start; j<=sz[1].end; j++) {
    for (i=sz[0].start; i<=sz[0].end; i++) {
	v[i + sz[0].ndim * j] = 
	    i + sz[0].gstart + (j + sz[1].gstart) * sz[0].mdim;
	}
    }

PIgsync( PIAllProcs );
t1 = SYGetElapsedTime();
PIWriteDistributedArray( fp, fmat, flen, sz, nd, v, MSG_DBL );
PIFclose( fp );
PIgsync( PIAllProcs );
t2 = SYGetElapsedTime() - t1;
if (PImytid == 0) {
    bytes = flen * nx * ny;
    printf( "Wrote %d bytes at %f Mbytes/second\n", bytes, bytes/(1.0e6*t2) );
    fflush( stdout );
    }

DBG("About to open" );
/* Now, try to READ the data back; check that it is correct */
fp = PIFopen( fname, PIAllProcs, O_RDONLY, pmode, -1 );
DBG("Opened file ");
if (!fp) {
    fprintf( stderr, "Can not open file for input\n" );
    return 0;
    }
if (rsize > 0 || wsize > 0) 
    PIFCacheEnable( fp, rsize, wsize );
DBG("Allocateing v2");
v2 = (double *)MALLOC( sz[0].ndim * sz[1].ndim * sizeof(double) );  CHKPTR(v2);

DBG("About to read common");
PIReadCommon( fp, "%4d",   4, &nxin, 1, MSG_INT );
PIReadCommon( fp, "%4d",   4, &nyin, 1, MSG_INT );
PIReadCommon( fp, "%4d\n", 5, &gin,  1, MSG_INT );
if (!fmat) {
    DBG("reading EOL(?!)");
    PIReadCommon( fp, "  \n", 3, (void *)0, 1, MSG_INT );
    }

DBG("About to gsync before read distributed");
PIgsync( PIAllProcs );
t1 = SYGetElapsedTime();
PIReadDistributedArray( fp, frmat, flen, sz, nd, v2, MSG_DBL );
PIFclose( fp );
PIgsync( PIAllProcs );
t2 = SYGetElapsedTime() - t1;
if (PImytid == 0) {
    bytes = flen * nx * ny;
    printf( "Read %d bytes at %f Mbytes/second\n", bytes, bytes/(1.0e6*t2) );
    fflush( stdout );
    }

/* Check that the values are correct */
if (nxin != nx) 
    fprintf( stderr, "[%d] NX is inccorect in file (%d should be %d)\n", 
	     PImytid, nxin, nx );
if (nyin != ny) 
    fprintf( stderr, "[%d] NY is inccorect in file (%d should be %d)\n", 
	     PImytid, nyin, ny );
if (gin != g) 
    fprintf( stderr, "[%d] G is inccorect in file (%d should be %d\n", 
	     PImytid, gin, g );
    
err = 0;
for (j=sz[1].start; j<=sz[1].end; j++) {
    int k;
    for (i=sz[0].start; i<=sz[0].end; i++) {
	k = i + sz[0].ndim * j;
	if (v[k] != v2[k]) {
	    fprintf( stderr, 
		   "[%d]Error in V values in file in location %d (%e != %e)\n",
		     PImytid, k, v[k], v2[k] );
	    err++;
	    if (err > 10) {
		fprintf( stderr, "Error number exceeded; exiting\n" );
		return err;
		}
	    }
	}
    }


FREE( v );
FREE( v2 );
return 0;
}
