#ifndef lint
static char SCCSid[] = "@(#) ./dgraph/citgraf.c 12/01/93";
#endif

/* C.It version */
#include <stdio.h>
#include "tools.h"
#include "dgraph/dgraph.h"

typedef struct {
    FILE *fp;
    int  use_color;
    } GRCitCtx;

extern void GRCitDrawLine(), GRCitAddLine(), GRCitDrawLines(),
	    GRCitStartPage(), GRCitEndPage(), GRCitDrawRectangle(),
	    GRCitDrawText(), GRCitClose(), GRCitDrawAxis(),
            GRCitSetLimits(), GRCitDrawHistogram();

/* Internal routines for cit */
char *GRCitiColorName( id )
int id;
{
switch (id) {
    case GRRED:   return "red";
    case GRGREEN: return "green";
    case GRBLUE:  return "blue";
    case GRYELLOW:return "yellow";
    case GRBLACK: return "black";
    default:
    	          return "black";
    }
}
char *GRCitiLineStyle( id )
int id;
{
switch (id) {
    case GRSOLID: return "solid";
    case GRDASH:  return "dash";
    case GRDOTS:  return "dots";
    default:
        return "";
    }
} 	          
void GRCitiSetColor( cctx, style )
GRCitCtx *cctx;
GRStyle  *style;
{
if (cctx->use_color) fprintf( cctx->fp, "set color %s\n",
				GRCitiColorName( style->color ) );
}
void GRCitiJoin( cctx, style )
GRCitCtx *cctx;
GRStyle  *style;
{
if (cctx->use_color) fputs( "join\n", cctx->fp );
else                 fprintf( cctx->fp, "join %s\n",
			      GRCitiLineStyle( style->bw_line_style ) );
}

/* General routines */
GRctx *GRCitCreate( fname, use_color )
char *fname;
int  use_color;
{
GRctx    *new;
GRCitCtx *cctx;

new = NEW(GRctx);                   CHKPTRN(new);
new->Dev = cctx = NEW(GRCitCtx);    CHKPTRN(new->Dev);
if (fname) {
    cctx->fp = fopen( fname, "w" ); CHKPTRN(cctx->fp);
    }
else
    cctx->fp = stdout;
    
new->use_color     = use_color;
new->is_interactive= 0;
new->DrawAxis      = GRCitDrawAxis;
new->SetLimits     = GRCitSetLimits;
new->DrawLine      = GRCitDrawLine;
new->AddLine       = GRCitAddLine;
new->DrawLines     = GRCitDrawLines;
new->StartPage     = GRCitStartPage;
new->EndPage       = GRCitEndPage;
new->DrawRectangle = GRCitDrawRectangle;
new->DrawText      = GRCitDrawText;
new->DrawHistogram = GRCitDrawHistogram;
new->Close         = GRCitClose;

return new;
}

void GRCitClose( gctx )
GRctx *gctx;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;
if (cctx->fp != stdout) 
    fclose( cctx->fp );
}
	
/* We let C.It manage the coordinate system */    
void GRCitDrawLine( gctx, x1, y1, x2, y2, style )
GRctx   *gctx;
double  x1, y1, x2, y2;
GRStyle *style;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;	
GRCitiSetColor( cctx, style );
fprintf( cctx->fp, "%e %e\n%e %e\n", x1, y1, x2, y2 );
GRCitiJoin( cctx, style );
}

void GRCitAddLine( gctx, x1, y1 )
GRctx  *gctx;
double x1, y1;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;	
fprintf( cctx->fp, "%e %e\n", x1, y1 );
}

void GRCitDrawLines( gctx, n, x, y, style )
GRctx   *gctx;
int     n;
double  *x, *y;
GRStyle *style;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;
int      i;

GRCitiSetColor( cctx, style );
for (i=0; i<n; i++) {
    fprintf( cctx->fp, "%e %e\n", x[i], y[i] );
    }
GRCitiJoin( cctx, style );
}

void GRCitStartPage( gctx )
GRctx  *gctx;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;

fprintf( cctx->fp, "set default\nset font variable\n" );
/* ? set limits ? */
}

void GRCitEndPage( gctx )
GRctx  *gctx;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;

fprintf( cctx->fp, "wait\nnew page\n" );
}

void GRCitDrawRectangle( gctx, x1, x2, y1, y2, style )
GRctx   *gctx;
double  x1, x2, y1, y2;
GRStyle *style;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;

GRCitiSetColor( cctx, style );
fprintf( cctx->fp, "%e %e\n%e %e\n%e %e\n%e %e\n%e %e\n",
     	 x1, y1, x1, y2, x2, y2, x2, y1, x1, y1 );
GRCitiJoin( cctx, style );
}

void GRCitDrawText( gctx, x, y, str, size, just )
GRctx  *gctx;
double x, y;
char   *str;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;

fprintf( cctx->fp, "text x %e y %e '%s'\n", x, y, str );
}

void GRCitSetLimits( gctx, xmin, xmax, ymin, ymax )
GRctx  *gctx;
double xmin, xmax, ymin, ymax;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;

gctx->xmin = xmin;
gctx->xmax = xmax;
gctx->ymin = ymin;
gctx->ymax = ymax;

fprintf( cctx->fp, "set limits x %e %e y %e %e\n", xmin, xmax, ymin, ymax );
}

void GRCitDrawAxis( gctx, flag )
GRctx  *gctx;
int    flag;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;

if (!flag) 
    fprintf( cctx->fp, "axis all off\n" );
}

void GRCitDrawHistogram( gctx, x, nx, binsize )
GRctx  *gctx;
int    nx;
double *x, binsize;
{
GRCitCtx *cctx = (GRCitCtx *)gctx->Dev;
int      i;

for (i=0; i<nx; i++) 
    fprintf( cctx->fp, "%e\n", x[i] );
fprintf( cctx->fp, "histogram binsize %f\n", binsize );
}

